from enum import Enum, auto


class RTL(Enum):
    VHDL = auto()
    VERILOG = auto()
    SYSTEM_VERILOG = auto()


class StopStep(Enum):
    SYNTHESIS = auto()
    IMPLEMENTATION = auto()


class ImplementationStep(Enum):
    PLACE = auto()
    ROUTE = auto()


class RegressionModel(Enum):
    KERNEL_RIDGE = auto()
    STREAM_LEARNING = auto()




import re
from pathlib import Path
from importlib_resources import files
from typing import Dict, Any, List, Optional
from bs4 import BeautifulSoup
from bs4.element import PageElement, ResultSet
import numpy as np


def _represents_float(s: Any):
    try:
        float(s)
        return True
    except ValueError:
        return False


def _get_percentage_util(columns: ResultSet, row: PageElement) -> str:
    column_number = 0
    column: PageElement
    for column in columns:
        if str(column.get("contents")) == "Util%":
            column_number: int = int(columns.index(column))
            break
    siebling = row
    if column_number > 0:
        for _ in range(0, column_number - 1):
            siebling: Any = siebling.findNext("tablecell")
    return siebling.findNext("tablecell").get("contents")


def get_available_indices(report_path: str) -> Dict[str, List[str]]:
    report: BeautifulSoup = BeautifulSoup(
        Path(report_path).open(), "lxml-xml",
    )
    indices = {}
    sections_partial: ResultSet = report.find_all("section")
    sections = []
    sections.extend(sections_partial)
    for section in sections:

        for nested_section in section.find_all("section"):
            nested_section.decompose()

        columns = []
        rows = []

        for column in (
            section.find("tablerow").find_all("tableheader")
            if section.find("tablerow")
            else []
        ):
            columns.append(column.get("contents").strip())
        if "Util%" in columns:
            for row in section.find_all("tablerow"):
                if row.find("tablecell") and _represents_float(
                    _get_percentage_util(section.find_all("tableheader"), row)
                ):
                    rows.append(row.find("tablecell").get("contents").strip())
            indices[section.get("title")] = rows

    return indices


def get_utilisation(report_path: str, section_name: str, row_name: str):
    report: BeautifulSoup = BeautifulSoup(
        Path(report_path).open(), "lxml-xml",
    )
    section: PageElement = report.find("section", {"title": section_name})
    columns: ResultSet = section.find_all("tableheader")
    row: PageElement = section.find(
        "tablecell",
        {"contents": re.compile("[ \t\r\n]*" + row_name + r"(\*)?[ \t\r\n]*")},
    )
    return float(_get_percentage_util(columns, row))


def get_wns(report_path: str) -> float:
    try:
        first_line = Path(report_path).open().readline()
        try:
            return float(
                re.findall(r"([-+]?(\d*\.\d*|inf))", first_line)[0][0]
            )
        except Exception:
            return float(np.inf)

    except FileNotFoundError as f:
        raise FileNotFoundError(
            "Wrong path: "
            + report_path
            + "\n"
            + "FileNotFoundError: "
            + str(f)
        )



from typing import List, Tuple
from collections import OrderedDict
from pymoo.model.problem import Problem
from pymoo.algorithms.nsga2 import NSGA2
from pymoo.factory import (
    get_sampling,
    get_crossover,
    get_mutation,
    get_termination,
)
from pymoo.optimize import minimize


import numpy as np
from dovado_rtl.simple_types import Metric
from dovado_rtl.fitness import FitnessEvaluator
import pickle


class MyProblem(Problem):
    def __init__(
        self,
        fitness_evaluator: FitnessEvaluator,
        free_parameters_range: "OrderedDict[str, Tuple[int, int]]",
        metrics: List[Metric],
    ):
        self.evaluator: FitnessEvaluator = fitness_evaluator
        self.metrics: List[Metric] = metrics
        super().__init__(
            n_var=len(free_parameters_range.keys()),
            n_obj=len(free_parameters_range.keys()),
            n_constr=0,
            xl=[
                free_parameters_range[parameter][0]
                for parameter in free_parameters_range.keys()
            ],
            xu=[
                free_parameters_range[parameter][1]
                for parameter in free_parameters_range.keys()
            ],
            type_var=np.int,
            elementwise_evaluation=True,
        )

    def _evaluate(self, x: List[int], out, *args, **kwargs):
        out["F"] = np.column_stack(
            [
                self.evaluator.fitness(tuple(x), metric)
                for metric in self.metrics
            ]
        )


def optimize(
    evaluator: FitnessEvaluator,
    free_parameters_range: "OrderedDict[str, Tuple[int, int]]",
    metrics: List[Metric],
    execution_time: str,
) -> List[float]:
    problem = MyProblem(evaluator, free_parameters_range, metrics)

    algorithm = NSGA2(
        pop_size=10 * len(free_parameters_range.keys()),
        n_offsprings=10 * len(free_parameters_range.keys()),
        sampling=get_sampling("int_random"),
        crossover=get_crossover("int_sbx", prob=0.9, eta=15),
        mutation=get_mutation("int_pm", eta=20),
        eliminate_duplicates=True,
    )

    termination = get_termination("time", execution_time)
    res = minimize(
        problem,
        algorithm,
        termination,
        seed=1,
        save_history=True,
        verbose=True,
    )
    with open("objs.pkl", "wb") as f:
        pickle.dump([res, algorithm], f)
    # with open('objs.pkl', 'rb') as f:
    #     obj0, obj1, obj2 = pickle.load(f)
    return res.F[-1].tolist()
